/*
 *  colgcc.c
 *  Copyright (C) 1999, 2000 by Richard Dawe
 *
 *  Run gcc or g++, etc. & colours the output. I borrowed the idea from a
 *  Perl script called colorgcc.
 *
 *  ---
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <errno.h>
#include <conio.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include <signal.h>
#include <unistd.h>
#include <sys/types.h>
#include <regex.h>

#include "ini.h"

/* Program info */
const char COLGCC_NAME[]     = "colgcc";
const char COLGCC_VERSION[]  = "1.2";
const char COLGCC_YEAR[]     = "1999, 2000";
const char COLGCC_AUTHOR[]   = "Richard Dawe";

const char COLGCC_CONFIG[]   = "colgcc.cfg";
const char COLGCC_SHAREDIR[] = "share/colgcc";

/* Colour data structure */
typedef struct {
  int info;
  int warning;
  int error;
  int make;
} colours_t;

/* Default colours */
static colours_t colours = {
  CYAN,     /* info */
  YELLOW,   /* warning */
  LIGHTRED, /* error */
  LIGHTBLUE /* make */
};

/* Configuration parsing */
static int colour_str2int (const char *str);
static int parse_config (char *buf, size_t bufsiz);

/* Console I/O bizness */
static struct text_info ti;

static void save_text_colour (void);
static void restore_text_colour (void);

/* The colour output functions. */
static void colour_output (char *filename);
static void colour_line   (char *line);

static void version_message (void);
static void help_message (void);

/* Nice Perl-isms */
static void die (char *msg);
static void warn (char *msg);

/* --------
   - main -
   -------- */

int
main (int argc, char *argv[])
{
  char config_file[PATH_MAX];
  int do_parse_config = 1; /* Parse config by default */
  char temp_file[L_tmpnam];
  char command[16384]; /* Handle up to max transfer buffer */
  int firstarg = 1;    /* First arg to pass on             */
  int debug    = 0;    /* Don't debug by default           */
  int i, ret;
  char *p = NULL;

  /* Check that there are parameters */
  if (argc < 2) {
    help_message();
    return(255);
  }

  /* Parse CLI args */
  for (i = 1; i < argc; i++) {
    /* Is this the first non-option argument - '-' indicates this?
     * Special case '/?', since it's used as a help option. */
    if ((*argv[i] != '-') && (strcmp(argv[i], "/?") != 0)) {
      firstarg = i;
      break;
    }

    /* Support the '--' syntax for indicating there are no more options. */
    if (strcmp(argv[i], "--") == 0) {
      firstarg = i + 1;
      break;
    }

    if (   (strcmp(argv[i], "-d")      == 0)
	|| (strcmp(argv[i], "--debug") == 0) )
      debug = 1;

    if (   (strcmp(argv[i], "-h")      == 0)
        || (strcmp(argv[i], "--help")  == 0)
	|| (strcmp(argv[i], "/?")      == 0) ) { /* Be nice to DOS users */
      help_message();
      return(255);
    }

    if (   (strcmp(argv[i], "-nc") == 0)
	|| (strcmp(argv[i], "--no-config") == 0)
	|| (strcmp(argv[i], "--no-configuration") == 0)) {
      /* Ignore any configuration files */
      do_parse_config = 0;
    }
  }

  /* Parse the configuration */
  if (   do_parse_config
      && !parse_config(config_file, sizeof(config_file))
      && (errno != ENOENT)) {
    perror(COLGCC_NAME);
    fprintf(stderr, "Configuration file: %s\n", config_file);
    die("Unable to parse configuration file!");
  }

  if (debug)
    printf("Using configuration file '%s'\n", config_file);

  /* Set up temp files for stdout, stderr */
  tmpnam(temp_file);

  /* Forward-slashify the filenames */
  for (p = temp_file; *p != '\0'; p++) { if (*p == '\\') *p = '/'; }

  /* Set up the command for redir to run the compiler */
  sprintf(command, "redir -o %s -eo ", temp_file);

  for (i = firstarg; i < argc; i++) {
    strcat(command, " ");
    strcat(command, argv[i]);
  }

  if (debug) {
    p = getcwd(NULL, PATH_MAX);
    printf("[In %s] '%s'\n", p, command);
    free(p);
  }

  /* Run it. Set the signal handlers as advised by the info page for
   * spawn*(). */
  signal(SIGINT, SIG_IGN);
  ret = system(command);
  signal(SIGINT, SIG_DFL);

  if (ret == -1) {
    perror(COLGCC_NAME);
    die("Unable to spawn child process!");
  }

  /* Parse the output. Because both stderr and stdout were captured,
   * output stderr first, like "redir -eo <blah>" does. */
  atexit(restore_text_colour);
  save_text_colour();

  colour_output(temp_file);    

  /* Kill the temporary files */
  unlink(temp_file);

  /* Pass on return code */
  return(ret);
}

/* -----------------
   - colour_output -
   ----------------- */

static void
colour_output (char *filename)
{
  FILE *fp         = NULL;    
  char *p          = NULL;
  int   use_colour = 1; /* Default to colour */
  char  buffer[1024];

  /* Only output in colour if the output is a terminal. */
  if (!isatty(fileno(stdout)))
    use_colour = 0;
    
  /* Open the temporary file. */
  fp = fopen(filename, "rt");
  if (fp == NULL) {
    perror(COLGCC_NAME);
    warn("Unable to open temporary file!");
    return;
  }

  while (fgets(buffer, sizeof(buffer), fp) != NULL) {
    /* Strip the newline off */
    if (strrchr(buffer, '\n') == buffer + strlen(buffer) - 1)
      buffer[strlen(buffer) - 1] = '\0';

    /* Convert tabs to spaces */
    for (p = buffer; *p; p++) {
      if (*p == '\t')
	*p = ' ';
    }

    /* Now output it */
    if (use_colour)
      colour_line(buffer);
    else
      puts(buffer);
  }

  fclose(fp);
}

/* ---------------
   - colour_line -
   --------------- */

static void
colour_line (char *line)
{
  static regex_t reg_we, reg_info, reg_make;
  static int     reg_compiled = 0;
  char *p = NULL;
  int ret;

  /* These regexps were taken from colorgcc 1.3.2 and modified not to use
   * ?, * or + operators. */
  if (!reg_compiled) {
    ret  = regcomp(&reg_we, "^(.{1,}):([0-9]{1,}):(.{1,})$",
		   REG_EXTENDED | REG_NOSUB);

    ret |= regcomp(&reg_info, "^(.{1,}):(.{1,}):$",
		   REG_EXTENDED | REG_NOSUB);

    ret |= regcomp(&reg_make, "^make(.{0,})\\[([0-9]{1,})\\]:(.{1,})$",
		   REG_EXTENDED | REG_NOSUB);

    reg_compiled = (ret == 0) ? 1 : 0;
  }

  /* Match the line to <filename>:<line num.>:<message> format */
  if (regexec(&reg_we, line, 0, NULL, 0) == 0) {
    /* Find the (warning|error) message to determine the colour */
    p = strstr(line, ":");
    if (p != NULL) p++, p = strstr(p, ":");
    if (p != NULL) p++;

    if (strstr(p, "warning:") != NULL) {
      textcolor(colours.warning);
    } else {
      textcolor(colours.error);
    }
  } else if (regexec(&reg_info, line, 0, NULL, 0) == 0) {
    textcolor(colours.info);
  } else if (regexec(&reg_make, line, 0, NULL, 0) == 0) {
    textcolor(colours.make);
  }

  cprintf("%s", line);
  restore_text_colour();
  cprintf("\r\n");
}

/* -------------------
 * - version_message -
 * ------------------- */

static void
version_message (void)
{
  printf("%s %s, Copyright (C) %s by %s\n"
	 "Distributed under terms of the GNU GPL\n\n",
	 COLGCC_NAME, COLGCC_VERSION, COLGCC_YEAR, COLGCC_AUTHOR);
}

/* ----------------
   - help_message -
   ---------------- */

static void
help_message (void)
{
  version_message();

  printf(
"%s colours the output of gcc and other compilers.\n"
"\n"
"Syntax:\n"
"    %s [<colgcc options>] <Compiler> [<Compiler parameters>]\n"
"\n",
COLGCC_NAME, COLGCC_NAME);

  printf(
"Options:\n"
"    -h, --help, /? - Display this help\n"
"    -d, --debug    - Display debugging information\n"
"    -nc, --no-config, --no-configuration\n"
"                   - Do not parse any configuration files\n"
"    --             - Last option indicator; remaining arguments are data.\n"
"\n"
"Examples:\n"
"    <Compiler>         = gcc\n"
"    <Compiler options> = -o myprog myprog.c\n"
"\n");
}

/* -------
 * - die -
 * ------- */

static void
die (char *msg)
{
  fprintf(stderr, "Error: %s\n", msg);
  exit(255);
}

/* --------
 * - warn -
 * -------- */

static void
warn (char *msg)
{
  fprintf(stderr, "Error: %s\n", msg);
}

/* --------------------
 * - save_text_colour -
 * -------------------- */

static void
save_text_colour (void)
{
  gettextinfo(&ti);
}

/* -----------------------
 * - restore_text_colour -
 * ----------------------- */

static void
restore_text_colour (void)
{
  textattr(ti.normattr);
}

/* ------------------
 * - colour_str2int -
 * ------------------ */

static int
colour_str2int (const char *str)
{
  int ret = -1; /* Fail by default */

  if (stricmp(str, "BLACK") == 0) {
    ret = BLACK;
  } else if (stricmp(str, "BLUE") == 0) {
    ret = BLUE;
  } else if (stricmp(str, "GREEN") == 0) {
    ret = GREEN;
  } else if (stricmp(str, "CYAN") == 0) {
    ret = CYAN;
  } else if (stricmp(str, "RED") == 0) {
    ret = RED;
  } else if (stricmp(str, "MAGENTA") == 0) {
    ret = MAGENTA;
  } else if (stricmp(str, "BROWN") == 0) {
    ret = BROWN;
  } else if (   (stricmp(str, "LIGHTGRAY") == 0)
	     || (stricmp(str, "LIGHT-GRAY") == 0)
	     || (stricmp(str, "LIGHTGREY") == 0)
	     || (stricmp(str, "LIGHT-GREY") == 0)) {
    ret = LIGHTGRAY;
  } else if (   (stricmp(str, "DARKGRAY") == 0)
	     || (stricmp(str, "DARK-GRAY") == 0)
	     || (stricmp(str, "DARKGREY") == 0)
	     || (stricmp(str, "DARK-GREY") == 0)) {
    ret = DARKGRAY;
  } else if (   (stricmp(str, "LIGHTBLUE") == 0)
	     || (stricmp(str, "LIGHT-BLUE") == 0)) {
    ret = LIGHTBLUE;
  } else if (   (stricmp(str, "LIGHTGREEN") == 0)
	     || (stricmp(str, "LIGHT-GREEN") == 0)) {
    ret = LIGHTGREEN;
  } else if (   (stricmp(str, "LIGHTCYAN") == 0)
	     || (stricmp(str, "LIGHT-CYAN") == 0)) {
    ret = LIGHTCYAN;
  } else if (   (stricmp(str, "LIGHTRED") == 0)
	     || (stricmp(str, "LIGHT-RED") == 0)) {
    ret = LIGHTRED;
  } else if (   (stricmp(str, "LIGHTBLUE") == 0)
	     || (stricmp(str, "LIGHT-BLUE") == 0)) {
    ret = LIGHTBLUE;
  } else if (   (stricmp(str, "LIGHTMAGENTA") == 0)
	     || (stricmp(str, "LIGHT-MAGENTA") == 0)) {
    ret = LIGHTMAGENTA;
  } else if (stricmp(str, "YELLOW") == 0) {
    ret = YELLOW;
  } else if (stricmp(str, "WHITE") == 0) {
    ret = WHITE;
  }

  return(ret);
}

/* ----------------
 * - parse_config -
 * ---------------- */

/*
 * This function locates the configuration file and parses it. The
 * configuration file's name is placed in buf. Up to bufsize - 1 characters
 * are written, plus a terminating nul.
 *
 * On success, 1 is returned. On failure, 0 is returned and errno is set
 * appropriately.
 */

static int
parse_config (char *buf, size_t bufsize)
{
  char filename[PATH_MAX];
  char data[16]; /* Enough to store the name of a colour */
  int found = 0;

  memset(filename, 0, sizeof(filename));

  /*
   * Find the configuration file. Look in the following places:
   *
   * 1. the current working directory;
   * 2. the user's home directory;
   * 3. the share directory off the DJGPP directory.
   */

  /* 1. Current working directory */
  getwd(filename);
  strcat(filename, "/");
  strcat(filename, COLGCC_CONFIG);

  if (access(filename, R_OK) == 0)
    found = 1;

  /* 2. Home directory */
  if (!found && (getenv("HOME") != NULL)) {
    sprintf(filename, "%s/%s", getenv("HOME"), COLGCC_CONFIG);
    if (access(filename, R_OK) == 0)
      found = 1;
  }

  /* 3. DJGPP share directory */
  if (!found && (getenv("DJDIR") != NULL)) {
    sprintf(filename, "%s/%s/%s",
	    getenv("DJDIR"), COLGCC_SHAREDIR, COLGCC_CONFIG);
    if (access(filename, R_OK) == 0)
      found = 1;
  }

  /* Copy the filename into the buffer, if found, else just put a nul. */
  if (found) {
    if ((buf != NULL) && (bufsize > 0)) {
      strncpy(buf, filename, bufsize - 1);
      buf[bufsize - 1] = '\0';
    }
  } else {
    if ((buf != NULL) && (bufsize > 0)) {
      *buf = '\0';
    }
  }

  /* Not found */
  if (!found) {    
    errno = ENOENT;
    return(0);
  }

  /* Get the colours from the config */

  if (GetPrivateProfileString("colours", "info", "cyan",
			      data, sizeof(data), filename) <= 0) {
    errno = EINVAL;
    return(0);
  } else {
    colours.info = colour_str2int(data);
    if (colours.info == -1) {
      errno = EINVAL;
      return(0);
    }
  }

  if (GetPrivateProfileString("colours", "warning", "yellow",
			      data, sizeof(data), filename) <= 0) {
    errno = EINVAL;
    return(0);
  } else {
    colours.warning = colour_str2int(data);
    if (colours.warning == -1) {
      errno = EINVAL;
      return(0);
    }
  }

  if (GetPrivateProfileString("colours", "error", "lightred",
			      data, sizeof(data), filename) <= 0) {
    errno = EINVAL;
    return(0);
  } else {
    colours.error = colour_str2int(data);
    if (colours.error == -1) {
      errno = EINVAL;
      return(0);
    }
  }

  if (GetPrivateProfileString("colours", "make", "lightblue",
			      data, sizeof(data), filename) <= 0) {
    errno = EINVAL;
    return(0);
  } else {
    colours.make = colour_str2int(data);
    if (colours.make == -1) {
      errno = EINVAL;
      return(0);
    }
  }

  return(1);
}
